#lang "qb"

#if __FB_LANG__="qb"
  #cmdline "-gen gcc -O 3 -lib -x libdirectfb.a" 
  option explicit
  #define i8 __byte
  #define u8 __ubyte
  #define i16 __short
  #define u16 __ushort
  #define i32 __long
  #define u32 __ulong
  #define pu8 u8 __ptr
  #define with __with
  #define screenres __screenres
  #define screenptr __screenptr
  #define screenlock __screenlock
  #define screenunlock __screenunlock
  #define cuint __cuint
  #define cast __cast
  #define cptr __cptr
  #define ptr __ptr
  #define zstring __zstring
  #define asm __asm  
  #define sizeof __sizeof
  #define allocate __allocate
  #define callocate __callocate
  #define deallocate __deallocate
  #define reallocate __reallocate
  #define dir __dir
  #define shl __shl
  #define shr __shr
  #define cvshort __cvshort
  #define setmouse __setmouse
  #define getmouse __getmouse
  #define multikey __multikey
  #define screensync __screensync
  #define flip __flip
  #define screencontrol __screencontrol
  #define imagecreate __imagecreate
  #define imagedestroy __imagedestroy
  #define iif __iif
  #define screenevent __screenevent
  #define curdir __curdir
#else    
  #cmdline "-gen gcc -O 3 -lib -x libdirectfb.a" 
  #include "crt.bi"
  #define i8 byte
  #define u8 ubyte
  #define i16 short
  #define u16 ushort
  #define pu8 u8 ptr
  #define __sleep sleep  
#endif

type pixel as u8

#include "directqb.bi"
#include "fbgfx.bi"

#if __FB_LANG__="qb"
  '$DYNAMIC
  defint A-Z
  declare sub puts cdecl alias "puts" (pz as zstring ptr)    
  sub dbgprint( s as string )
    dim as zstring ptr p : p = strptr(s)
    asm    
      push [p]
      call puts
      add esp, 4
    end asm
  end sub
#else
  using fb
  sub dbgprint( s as string ) 
    puts(s)
  end sub
#endif

#define LogFunc rem
'#define LogFunc dbgprint

'----------------------------------------------------------------------------
' DIRECTQB.BI
'   Include file for the DirectQB library version 1.61
'   by Angelo Mottola - Enhanced Creations 1998-99
'
' This file contains constants and function declarations used by the library
' Always include this file into your own programs!
'
' Include file created on 11-10-1999 by the DirectQB Library Manager v1.2
' Run DQBMAN again to modify your DirectQB module settings
'
'  Read DIRECTQB.DOC for detailed informations on how to use the library
'----------------------------------------------------------------------------

#macro Unimplemented()
  dbgprint "Unimplemented: '" & __FUNCTION__ & "'" : do: sleep: loop
  '!!! works generically but bugs freebasic? !!!
  '#if __FB_QUERY_SYMBOL__( 1 , __FUNCTION_NQ__ )=17
  '  __FUNCTION_NQ__ = ""
  '#elseif __FB_QUERY_SYMBOL__( 1 , __FUNCTION_NQ__ )
  '  __FUNCTION_NQ__ = 0
  '#endif
#endmacro  
#macro UnimplementedN()
  Unimplemented()
  __FUNCTION_NQ__ = 0
#endmacro  
#macro UnimplementedS()
  Unimplemented()
  __FUNCTION_NQ__ = ""  
#endmacro  

type DQB_Mouse
  as i8  bLock, bPad
  as i16 XPos,YPos
  as i16 Wheel,Btn
  as i16 L,T,R,B
end type
type DQB_Context  
  bInitialized  as i8  
  iLayerCount   as u8
  iSoundCount   as u8
  bUpdatePal    as u8   
  iFPS          as i16
  iTexSize      as i16  
  tMouse        as DQB_Mouse
  pLayer(15)    as image ptr  
  pFontText     as image ptr
  pXms(255)     as any ptr  
end type

static shared as DQB_Context tDQB

'Option Nogosub
#define _return(_N) __FUNCTION_NQ__ = (_N) : exit function

sub _DQB_PumpEvents()
  dim as EVENT ev
  with tDQB.tMouse
    while screenevent(@ev)
      select case ev.type
      case EVENT_MOUSE_MOVE
        if .bLock then
          if ev.x <> 160 orelse ev.y <> 120 then setmouse 160,120
          .XPos += (ev.x-160)\2 : .YPos += (ev.y-120)\2
        else
          .Xpos = ev.x : .Ypos = ev.y
        end if
        if .XPos < .L then .XPos=.L
        if .YPos < .T then .YPos=.T
        if .XPos > .R then .XPos=.R
        if .YPos > .B then .YPos=.B
      case EVENT_MOUSE_BUTTON_PRESS
        if (ev.button = BUTTON_LEFT) Then
          .Btn or= 1
        elseif (ev.button = BUTTON_RIGHT) Then
          .Btn or= 2
        else
          .Btn or= 4
        end if
      case EVENT_MOUSE_BUTTON_RELEASE
        if (ev.button = BUTTON_LEFT) Then
          .Btn and= (not 1)
        elseif (ev.button = BUTTON_RIGHT) Then
          .Btn and= (not 2)
        else
          .Btn and= (not 4)
        end if
  
      
      case EVENT_WINDOW_GOT_FOCUS
        if .Block then setmouse ,,0,1 : setmouse 160,120
      case EVENT_WINDOW_LOST_FOCUS
        if .Block then setmouse ,,1,0
      end select
    wend
  end with
end sub

' Procedures from MAIN.OBJ:
function DQBinit (byval NumLayers as integer, byval NumSounds as integer, byval MemSize as integer) as integer 'OK
  DbgPrint("DQBInit!")
  LogFunc "DQBInit(NumLayers=" & NumLayers & " , NumSounds=" & NumSounds & " , MemSize=" & MemSize & ")"
  if tDQB.bInitialized then _return(4) 'Already Initialized
  tDQB.bInitialized = 1
  if NumLayers > 7 then NumLayers = 7
  tDQB.iLayerCount = NumLayers
  if NumSounds > 8 then NumSounds = 8
  tDQB.iSoundCount = NumSounds  
  tDQB.pLayer(0)=0
  dim as long i
  for i = 1 to NumLayers
    tDQB.pLayer(i) = callocate( 320*200l + sizeof(_OLD_HEADER) )        
    if tDQB.pLayer(i) = 0 then 
      DbgPrint("DBQInit: Failed to allocate layers")
      _return(3) 'failed to allocate layers
    end if
    with *tDqb.pLayer(i)                  
      .old.bpp = 8 : .old.width = 320 : .old.height = 200      
    end with
  next i
  _return(0) 'Success
end function
function DQBver () as integer
  UnimplementedN()
end function
function qbstr(DQBid) () fbstr
  UnimplementedS()  
end function
function DQBmapLayer (byval Layer  as integer) as integer
  UnimplementedN()
end function  
sub DQBclose () 'OK
  dim as long i
  for i = 1 to 9
    if tDQB.pLayer(i) then deallocate( tDQB.pLayer(i) ) : tDQB.pLayer(i)=0
  next i
end sub
sub DQBpeek (byval DataSeg as integer, byval DataOff as integer, byval Offset as long, byval Length as integer)
  Unimplemented()
end sub
sub DQBpoke (byval DataSeg as integer, byval DataOff as integer, byval Offset as long, byval Length as integer)
  Unimplemented()
end sub
sub DQBsort (byval ArraySeg as integer, byval ArrayOff as integer, byval NumRecords as integer, byval RecordLen as integer, byval IndexOff as integer)
  Unimplemented()
end sub
function DQBangle (byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer) as integer
  UnimplementedN()
end function  
sub DQBinitVGA () 'OK
  LogFunc "DQBInitVGA()"
  #ifdef __FB_DOS__
    screenres 320,200,8
    if screenptr=0 then screen 13    
  #else
    screencontrol(SET_GL_2D_MODE , OGL_2D_MANUAL_SYNC)
    screencontrol(SET_GL_SCALE , 3)
    screenres 320,200,8,,GFX_OPENGL    
  #endif
  if screenptr=0 then 
    DbgPrint("Failed to init VGA")
    sleep : end 1
  end if  
end sub
sub DQBinitText () 'OK
  'Unimplemented()
  screen 0
end sub
function DQBsetBaseLayer (byval Layer as integer) as integer 'OK?
  LogFunc "DQBsetBaseLayer(Layer=" & Layer & ")"  
  _return(0) 'already allocated
end function
sub DQBcopyLayer (byval SourceLayer as integer, byval DestLayer as integer) 'OK
  LogFunc "DQBcopyLayer(SourceLayer=" & SourceLayer & " , DestLayer=" & DestLayer & ")"
  if SourceLayer > 7 then dbgprint("Source Layer: "+hex$(SourceLayer))
  if DestLayer > 7 then dbgprint("DestLayer: "+hex$(SourceLayer))
  dim as image ptr pSrc,pDst
  pDst=tDQB.pLayer(  DestLayer and 7)
  pSrc=tDQB.pLayer(SourceLayer and 7)
  if pSrc=pDst then dbgprint("DQBcopyLayer: Invalid copy Src=" & SourceLayer & " Dst=" & DestLayer) : exit sub
  if pSrc=0 then dbgprint("DQBcopyLayer: source is screen Src=" & SourceLayer & " Dst=" & DestLayer)
  put pDst,(0,0),pSrc,pset    
  if (DestLayer=0) then 
    'if DQBframeReady() then flip
    flip
  end if
end sub
sub DQBclearLayer (byval Layer as integer) 'OK
  LogFunc "DQBclearLayer(Layer=" & Layer & ")"
  line tDQB.pLayer(Layer and 7),(0,0)-(319,199),0,bf
end sub
sub DQBwait (byval Times as integer) 'OK
  LogFunc "DQBwait(Times=" & Times & ")"  
  dim as long N
  if tDQB.bUpdatePal = 1 then
    tDQB.bUpdatePal = 0
    'dim as double dTemp : dTemp = __timer 
    flip : Times -= 1
    'if abs(timer-dTemp) > (1/120) then Times -= 1
    'timer
  end if
  for N = 0 to Times-1
    screensync    
  next N  
end sub
sub DQBsetFrameRate (byval FPS as integer) 'ok
  LogFunc "DQBsetFrameRate(FPS=" & FPS & ")"
  tDQB.iFPS = FPS  
end sub
function DQBframeReady () as integer 'ok
  LogFunc "=DQBframeReady()"
  static as double dTMR
  if abs(timer-dTMR) > 1 then dTMR = timer : _return(0)    
  if abs(timer-dTMR) > (1/tDQB.iFPS) then dTMR += (1/tDQB.iFPS) : _return(-1)
  _return(0)
end function
function qbstr(DQBerror) () fbstr
  UnimplementedS()
end function

' Procedures from DRAW.OBJ:
sub DQBcopyTransLayer (byval SourceLayer as integer, byval DestLayer as integer) 'OK
  LogFunc "DQBcopyTransLayer(SourceLayer=" & SourceLayer & " , DestLayer=" & DestLayer & ")"
  put tDQB.pLayer(DestLayer and 7),(0,0),tDQB.pLayer(SourceLayer and 7),trans
end sub
sub DQBcopyHitLayer (byval SourceLayer as integer, byval DestLayer as integer, byval Col as integer)
  Unimplemented()
end sub
sub DQBpset (byval Layer as integer, byval x as integer, byval y as integer, byval Col as integer) 'OK
  pset tDQB.pLayer(Layer and 7) , ( x,y ), col  
end sub
function DQBpoint(byval Layer as integer, byval x as integer, byval y as integer) as integer 'OK
  _return(point(x,y,tDQB.pLayer(Layer and 7)))
end function
sub DQBline (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval Col as integer)
  Unimplemented()
end sub
sub DQBgline (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval Col1 as integer, byval Col2 as integer)
  Unimplemented()
end sub
sub DQBellipse (byval Layer as integer, byval x as integer, byval y as integer, byval rx as integer, byval ry as integer, byval Col as integer)
  Unimplemented()
end sub
sub DQBbox (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval Col as integer)
  Unimplemented()
end sub
sub DQBboxf (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval Col as integer)
  Unimplemented()
end sub
sub DQBpaint (byval Layer as integer, byval x as integer, byval y as integer, byval Col as integer)
  Unimplemented()
end sub
sub DQBscroll (byval Layer as integer, byval dx as integer, byval dy as integer)
  Unimplemented()
end sub
sub DQBscrollArea (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval Direction as integer)
  Unimplemented()
end sub
sub DQBsetTransPut ()
  Unimplemented()
end sub
sub DQBsetSolidPut ()
  Unimplemented()
end sub
sub DQBget (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval DummyBufSeg as integer, byval pBuf as any ptr) 'OK
  logfunc "DQBget(Layer=" & Layer & " , " & "(" & x1 & "," & y1 & ")-(" & x2 & "," & y2 & "), pBuf=0x" & hex$(pBuf,8) & ")"
  if x1 > x2 then swap x1,x2
  if y1 > y2 then swap y1,y2
  dim as image ptr pImg   : pImg = pBuf
  dim as image ptr pLayer : pLayer = tDQB.pLayer(Layer and 7)  
  with pImg->old
    .bpp = 0 : .width = (x2-x1)+1 : .height = (y2-y1)+1
    put pImg,(0,0),pLayer,(x1,y1)-(x2,y2),pset
  end with
end sub
sub DQBput (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval pBuf as any ptr) 'OK
  logfunc "DQBput(Layer=" & Layer & " , " & "(" & x & "," & y & "), pBuf=0x" & hex$(pBuf,8) & ")"
  put tDQB.pLayer(Layer and 7),(x,y),pBuf,pset
end sub

' Procedures from IMAGE.OBJ:
function DQBloadImage (byval Layer as integer, byval x as integer, byval y as integer, FileName as string, Pal as string, byref imgWidth as integer, byref imgHeight as integer) as integer 'OK
  
  LogFunc "DQBloadImage(Layer=" & Layer & " , x=" & x & " , y=" & y & " , FileName=" & FileName & ")"
  
  type PcxHeader field = 1
    Magic        as u8  '10
    Version      as u8  '3,4,5?
    Encode       as u8  '0=raw , 1=rle
    Bpp          as u8  '1,2,4,(8)
    MinX         as i16 
    MinY         as i16 
    MaxX         as i16 
    MaxY         as i16 
    HorzDPI      as i16
    VertDPI      as i16
    EgaPal(47)   as u8  'EGA Pal
    Reserved0    as u8
    Planes       as u8  '(1) 3, 4
    Pitch        as i16
    PalMode      as i16 '1/2 ???
    HorzRes      as i16
    VertRes      as i16
    Reserved(53) as u8
  end type
  
  dim pScr as pu8 , bLock as u8 : pScr = cptr(pu8,tDQB.pLayer(Layer and 7))
  if pScr=0 then pScr = screenptr : bLock = 1 else pScr += sizeof(_OLD_HEADER)
  dim as integer f : f = freefile()
  if open(FileName for binary access read as #f) then dbgPrint("DQBloadImage failed to open '"+FileName+"'"): _return( 1 ) 'File couldnt be opened
  dim as u16 uMagic : get #f,,uMagic
  select case uMagic
  case &h030A,&h040A,&h050A 'PCX v3-5
    dim as PcxHeader tHdr : get #f,1,tHdr
    #define iLeft x 
    #define iTop y
    const iWidth=320,iHeight=200 'dqb layer size
    with tHdr  
      'if .Magic  <> 10 then dbgprint "Not a .pcx file" : sleep : system
      if .Bpp    <> 8  then dbgprint "PCX: Unsupported bpp: " & .Bpp : close #f : _return(3)
      if .Planes <> 1  then dbgprint "PCX: Unsupported planes: " & .Planes : close #f : _return(3)
      if .Pitch <> ((.MaxX-.MinX)+1) then dbgprint "PCX: Warning Pitch differ: " & .Pitch          
      dim as long iPos,uAmount,N,iSrcX,iSrcY,iX,iY
      dim as u8 bBuffer(4095+768),uTmp,uCor
      uAmount = -1 : iY = iTop
      ImgWidth = .Pitch : ImgHeight=((.MaxY-.MinY)+1)      
      if iY < cuint(iHeight) then iX = iLeft else iX = iWidth  
      if bLock then screenlock  
      do
        #macro WritePixel(_Pix)
          if cuint(iX) < iWidth then pScr[iX] = _Pix
          iX += 1 : iSrcX += 1
          if iSrcX = ImgWidth then 
            iSrcX = 0 : pScr += iWidth : iY += 1 : iSrcY += 1          
            if iSrcY = ImgHeight then exit do
            if iY < cuint(iHeight) then iX = iLeft else iX = iWidth  
          end if
        #endmacro
        if iPos >= uAmount then
          uAmount = (uAmount+1)-iPos 'remaining
          if uAmount = 1 then bBuffer(0)=bBuffer(iPos)
          get #f,,bBuffer(uAmount),4096-uAmount,iPos
          uAmount += (iPos-1) : iPos = 0
        end if    
        uTmp = bBuffer(iPos) : iPos += 1
        if uTmp >= 192 then      
          uCor = bBuffer(iPos) : iPos += 1      
          for N = 193 to uTmp
            WritePixel(uCor)
          next N
        else
          WritePixel( uTmp )      
        end if
      loop
      uTmp = bBuffer(iPos) : iPos += 1 : uAmount += 1
      if uTmp <> 12 then get #f,lof(f)-768,uTmp : uAmount=0 : iPos=0
      if uTmp = 12 then    
        get #f,,bBuffer(uAmount),768-(uAmount-iPos)    
        Pal = space$(768)        
        for N = 0 to 767
          Pal[N] = bBuffer(iPos+N) shr 2 '0 to 63
        next N        
      else
        DbgPrint("PCX: Warning: no palette found")
      end if      
      if bLock then screenunlock  
    end with
    close #f: _return(0)
  case cvshort("BM")        'BMP
    close #f : dbgPrint("TODO: implement BMP"): _return(2)
  case &h0006               'BSV
    close #f : dbgPrint("TODO: implement BSV"): _return(2)
  case else                 'Bad/Unknown file format
    close #f : dbgPrint("Unknown image format"):_return(3)
  end select  
end function
function DQBsaveImage (Layer as integer, x1 as integer, y1 as integer, x2 as integer, y2 as integer, FileName as string, Pal as string, Format as integer) as integer
  UnimplementedN()
end function
function DQBplayFLI (FileName as string, BufLayer as integer, KeyStop as integer, LoopFlag as integer) as integer
  UnimplementedN()
end function
function DQBopenFLI (FileName as string, Frames as integer, Speed as integer) as integer
  UnimplementedN()
end function
sub DQBplayFLIstep (byval Layer as integer)
  Unimplemented()
end sub
sub DQBcloseFLI ()
  Unimplemented()
end sub

' Procedures from SPRITE.OBJ: 
function DQBsize (byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer) as integer
  UnimplementedN()
end function
sub DQBsetClipBox (byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer)
  Unimplemented()
end sub
sub DQBfPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr)
  Unimplemented()
end sub
sub DQBsPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval NewWidth as integer, byval NewHeight as integer)
  Unimplemented()
end sub
sub DQBrPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval Angle as integer, byval Zoom as integer) 'needs to rotate/zoom
  const PI256 = (atn(1)*4)/128  
  dim as image ptr pBuf : pBuf = BufOff
  dim as long iWid,iHei : iWid = pBuf->Old.Width : iHei = pBuf->Old.Height  
  dim as single fWH,fHH,fScale : fWH = iWid/2 : fHH = iHei/2 : fScale = Zoom/100
  dim as long iWidScale,iHeiScale : iWidScale = iWid*fScale : iHeiScale = iHei*fScale  
  
  dim as single HX,HY : HX=cos(PI256*-Angle)/fScale      : HY=sin(PI256*-Angle)/fScale
  dim as single VX,VY : VX=cos(PI256*(-Angle+64))/fScale : VY=sin(PI256*(-Angle+64))/fScale
  dim as single FX,FY , fLeft,fTop,fRight,fBottom
  #macro ChkLimit()
    if FX < fLeft   then fLeft   = FX
    if FX > fRight  then fRight  = FX
    if FY < fTop    then fTop    = fY
    if FY > fBottom then fBottom = fY
  #endmacro
  FX = HX*(iWidScale) : FY = HY*(iHeiScale)
  ChkLimit()
  FX = VX*(iWidScale) : FY = VY*(iHeiScale)
  ChkLimit()
  FX = (VX+HX)*(iWidScale) : FY = (VY+HY)*(iHeiScale)
  ChkLimit()  
  FX = fWH-((fLeft+fRight+.5)/2) : FY = fHH-((fTop+fBottom+.5)/2)
  'select case Angle
  'case  64 to 127 : FY -= .5
  'case 128 to 191 : FX -= .5 : FY -= .33
  'case 192 to 255 : FX -= .33 : FY += .33
  'end select  
  dim as u8 ptr pRow 
  if (Layer and 7) then
    pRow=cptr(u8 ptr,tDQB.pLayer(Layer and 7)) : if pRow=0 then exit sub
    pRow += sizeof(_OLD_HEADER)+y*320+x
  else
    screenlock : pRow=screenptr+y*320+x
  end if
  
  dim as u8 ptr pSrc : pSrc=BufOff+sizeof(_OLD_HEADER)  
  dim as single PX,PY
  dim as long nX,nY,iX,iY
  dim as u8 bPix
  for nY = 0 to iHeiScale-1
    PX=FX : PY=FY    
    for nX = 0 to iWidScale-1
      iX = int(PX) : iY = int(PY)
      if cuint(iX)<iWid andalso cuint(iY)<iHei then
        bPix = pSrc[iY*iWid+iX]
        if bPix then pRow[nX] = bPix
      end if
      PX += HX : PY += HY
    next nX
    FX += VX : FY += VY : pRow += 320
  next nY
  if (Layer and 7)=0 then screenunlock  
end sub
sub DQBxPut (byval SourceLayer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval DestLayer as integer, byval x as integer, byval y as integer)
  Unimplemented()
end sub
sub DQBmPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval iflip as integer)
  Unimplemented()
end sub
sub DQBhPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval Col as integer)
  Unimplemented()
end sub
sub DQBtPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval BitMode as integer)
  Unimplemented()
end sub
sub DQBpPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval Pattern as integer)
  Unimplemented()
end sub
sub DQBputOver (byval DummyBackSeg as integer, byval BackOff as any ptr, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr)
  Unimplemented()
end sub
sub DQBsetCollideMethod (byval Method as integer)
  Unimplemented()
end sub
function DQBcollide (byval x1 as integer, byval y1 as integer, byval DummySeg1 as integer, byval Off1 as any ptr, byval x2 as integer, byval y2 as integer, byval DummySeg2 as integer, byval Off2 as any ptr) as integer
  UnimplementedN()
end function
function DQBcollideOnLayer (byval Layer as integer, byval x as integer, byval y as integer, byval DummySpriteSeg as integer, byval SpriteOff as any ptr) as integer
  UnimplementedN()
end function

' Procedures from BIT.OBJ:
function DQBsetBit (byval Value as integer, byval bbit as integer) as integer
  UnimplementedN()
end function
function DQBresetBit (byval Value as integer, byval bbit as integer) as integer
  UnimplementedN()
end function
function DQBreadBit (byval Value as integer, byval bbit as integer) as integer
  UnimplementedN()
end function
function DQBtoggleBit (byval Value as integer, byval bbit as integer) as integer
  UnimplementedN()
end function
function DQBshiftLeft (byval Value as integer, byval nBits as integer) as integer
  UnimplementedN()
end function
function DQBshiftRight (byval Value as integer, byval nBits as integer) as integer
  UnimplementedN()
end function

' Procedures from PALETTE.OBJ:
sub DQBsetCol (byval ColorIndex as integer, byval Red as integer, byval Green as integer, byval Blue as integer) 'OK
  palette ColorIndex, (Red*255)\63,(Green*255)\63,(Blue*255)\63  
end sub
sub DQBgetCol (byval ColorIndex as integer, Red as integer, Green as integer, Blue as integer) 'OK
  dim as long R,G,B
  palette get ColorIndex, R,G,B
  Red = R shr 2 : Green = G shr 2 : Blue = B shr 2
end sub
function DQBfindCol (byval Red as integer, byval Green as integer, byval Blue as integer) as integer
  UnimplementedN()
end function
function DQBfindPalCol (Pal as string, Red as integer, Green as integer, Blue as integer) as integer
  UnimplementedN()
end function
sub DQBsetPal (Pal as string) 'OK
  LogFunc "DQBSetPal(Pal[])"
  dim as long N,I
  for N = 0 to (len(Pal)\3)-1
    palette N,(Pal[I+0] shl 0l) or (Pal[I+1] shl 8l) or (Pal[I+2] shl 16l) : I += 3
  next N  
  tDQB.bUpdatePal = 1
end sub
sub DQBgetPal (Pal as string)
  Unimplemented()
end sub
sub DQBfadeIn (Pal as string)
  Unimplemented()
end sub
sub DQBfadeStepIn (Pal as string) 'OK
  dim as long N,R,G,B,I,Red,Green,Blue
  for N=0 to 255
    palette get N,R,G,B : R shr= 2 : G shr= 2 : B shr= 2 
    Red = Pal[I] : Green = Pal[I+1] : Blue = Pal[I+2]: I += 3
    R += sgn(Red-R) : G += sgn(Green-G) : B += sgn(Blue-B)    
    palette N,(R*255)\63,(G*255)\63,(B*255)\63
  next N 
  flip
end sub
sub DQBfadeTo (byval Red as integer, byval Green as integer, byval Blue as integer)
  Unimplemented()
end sub
sub DQBfadeStepTo (byval Red as integer, byval Green as integer, byval Blue as integer) 'OK
  dim as long N,R,G,B
  for N=0 to 255
    palette get N,R,G,B : R shr= 2 : G shr= 2 : B shr= 2 
    R += sgn(Red-R) : G += sgn(Green-G) : B += sgn(Blue-B)    
    palette N,(R*255)\63,(G*255)\63,(B*255)\63
  next N   
  tDQB.bUpdatePal = 1
end sub
sub DQBpalOff ()
  Unimplemented()
end sub
sub DQBpalRotate (byval FirstCol as integer, byval LastCol as integer, byval RotDir as integer)
  Unimplemented()
end sub

enum TextStyle
  tsNONE       =  0 
  tsSOLID      =  1
  tsBOLD       =  2
  tsITALIC     =  4
  tsUNDERLINED =  8
  tsBLENDED    = 16
  tsTEXTURED   = 32 
end enum

' Procedures from FONT.OBJ:
sub DQBprint (Layer as integer, Text as string, x as integer, y as integer, Col as integer)
  Unimplemented()
end sub
sub DQBprints (byval Layer as integer, Text as string, byval x as integer, byval y as integer, byval Col as integer, byval Style as integer) 'OK
  dim as image ptr pLayer : pLayer = tDQB.pLayer(Layer and 7)
  dim as long N,iBold,PX,iLen : iBold = (Style and BOLD)<>0 : iLen = len(Text)
  'if (Style and BLENDED) then DbgPrint "DQBprints: BLENDED style not implemented"
  'if (Style and ITALIC) then DbgPrint "DQBprints: ITALIC style not implemented"  
      
  if (Style and ITALIC) then 
    dim as image ptr pTemp : pTemp=ImageCreate(iLen*8,8)
    'if (Style and SOLID) andalso pLayer<>0 then line pTemp,(0,0)-step((not iBold)+(iLen*8),7),0,bf
    if (Style and TEXTURED) then
      Col = 0 : PX = 0 : for N = 0 to iLen-1 : put pTemp,(PX,0),tDQB.pFontText,xor : PX += 8 : next N
    end if  
    draw string pTemp,(0,0),Text,Col
    if iBold then draw string pTemp,(1,0),Text,Col
    if (Style and UNDERLINED) then line pTemp,(0,7)-step((not iBold)+(iLen*8),0),Col
    if (Style and TEXTURED) then
      PX = 0 : for N = 0 to iLen-1 : put pTemp,(PX,0),tDQB.pFontText,xor : PX += 8 : next N
    end if
    if (Style and SOLID) then
      put pLayer,(x-2,y  ),pTemp,(0,0)-(iLen*8-1,2),pset
      put pLayer,(x-1,y+2),pTemp,(0,3)-(iLen*8-1,4),pset
      put pLayer,(x  ,y+4),pTemp,(0,5)-(iLen*8-1,7),pset
    else
      put pLayer,(x-2,y  ),pTemp,(0,0)-(iLen*8-1,2),trans
      put pLayer,(x-1,y+2),pTemp,(0,3)-(iLen*8-1,4),trans
      put pLayer,(x  ,y+4),pTemp,(0,5)-(iLen*8-1,7),trans
    end if
    ImageDestroy(pTemp)
  else
    'if (Style and SOLID) andalso pLayer<>0 then line pLayer,(x,y)-step((not iBold)+(iLen*8),7),0,bf
    if (Style and TEXTURED) then
      Col = 0 : PX = x : for N = 0 to iLen-1 : put pLayer,(PX,y),tDQB.pFontText,xor : PX += 8 : next N
    end if  
    draw string pLayer,(x,y),Text,Col
    if iBold then draw string pLayer,(x+1,y),Text,Col
    if (Style and UNDERLINED) then line pLayer,(x,y+7)-step((not iBold)+(iLen*8),0),Col
    if (Style and TEXTURED) then
      PX = x : for N = 0 to iLen-1 : put pLayer,(PX,y),tDQB.pFontText,xor : PX += 8 : next N
    end if
  end if
  
end sub
function DQBlen (Text as string) as integer
  _return(len(Text))
end function
sub DQBsetBIOSfont ()
  Unimplemented()
end sub
function DQBloadFont (FileName as string) as integer
  UnimplementedN()
end function
sub DQBsetFont (Font as string)
  Unimplemented()
end sub
sub DQBsetTextStyle (byval Style as integer)
  Unimplemented()
end sub
sub DQBsetTextBackCol (byval Col as integer)
  Unimplemented()
end sub
sub DQBsetTextSpacing (byval Spacing as integer)
  Unimplemented()
end sub
sub DQBsetTextBMap (byval BMap as integer)
  Unimplemented()
end sub
sub DQBsetFontTexture (byval DummyTextSeg as integer, byval TextOff as any ptr) 'OK
  tDQB.pFontText = TextOff
end sub

' Procedures from DISK.OBJ:
function qbstr(DQBdir) (Mask as string, Attributes as integer) fbstr 'OK
  _return( dir(Mask,Attributes) )
end function
function qbstr(DQBdrive) () fbstr 'OK
  _return(left$(curdir(),1))
end function
function qbstr(DQBpath) () fbstr 'OK
  _return(curdir())
end function
function DQBnumDrives () as integer 'OK
  UnimplementedN()
end function
sub DQBsetDrive (NewDrive as string) 'OK
  chdir(chr$(asc(NewDrive,1),asc(":")))
end sub
function DQBchDir (NewDir as string) as integer 'OK
  _return(chdir(NewDir))
end function

' Procedures from BLENDING.OBJ:
sub DQBfilterBox (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval Col as integer, byval BMap as integer)
  Unimplemented()
end sub
sub DQBbPut (byval Layer as integer, byval x as integer, byval y as integer, byval DummyBufSeg as integer, byval BufOff as any ptr, byval BMap as integer)
  Unimplemented()
end sub
sub DQBcopyBlendLayer (byval SourceLayer as integer, byval DestLayer as integer, byval BMap as integer)
  Unimplemented()
end sub
function DQBcreateBMap (byval BMap as integer, byval FirstCol as integer, byval LastCol as integer) as integer
  UnimplementedN()
end function
function DQBloadBMap (BMap as integer, FileName as string) as integer
  UnimplementedN()
end function
function DQBsaveBMap (BMap as integer, FileName as string) as integer
  UnimplementedN()
end function
sub DQBsetBMap (byval BMap as integer, byval ForeCol as integer, byval BackCol as integer, byval NewCol as integer)
  Unimplemented()
end sub
function DQBgetBMap (byval BMap as integer, byval ForeCol as integer, byval BackCol as integer) as integer
  UnimplementedN()
end function

' Procedures from KEYBOARD.OBJ:
sub DQBinstallKeyboard () 'OK
  rem nothing to do
end sub
sub DQBremoveKeyboard () 'OK
  rem nothing to do
end sub
function DQBkey (byval ScanCode as integer) as integer
  _return(multikey(ScanCode))
end function
function DQBreadKey () as integer
  UnimplementedN()
end function
sub DQBwaitKey (byval ScanCode as integer)
  Unimplemented()
end sub
function DQBasc (byval ScanCode as integer, byval ShiftFlag as integer) as integer
  UnimplementedN()
end function
function qbstr(DQBinkey) () fbstr
  UnimplementedS()
end function

' Procedures from JOYSTICK.OBJ:
function DQBjoyDetected (byval JoyNum as integer) as integer
  UnimplementedN()
end function
function DQBjoyX (byval JoyNum as integer) as integer
  UnimplementedN()
end function
function DQBjoyY (byval JoyNum as integer) as integer
  UnimplementedN()
end function
function DQBjoyMove (byval JoyNum as integer, byval Direction as integer) as integer
  UnimplementedN()
end function
function DQBjoyFire (byval JoyNum as integer, byval Button as integer) as integer
  UnimplementedN()
end function
sub DQBpollJoy (byval JoyNum as integer)
  Unimplemented()
end sub
sub DQBresetJoy ()
  Unimplemented()
end sub
sub DQBsetJoySens (byval Sens as integer)
  Unimplemented()
end sub

' Procedures from MOUSE.OBJ:
function DQBmouseDetected () as integer 'OK
  _return(1) 'Mouse Detected
end function
function DQBmouseX () as integer 'OK
  _DQB_PumpEvents()  
  _return( tDQB.tMouse.XPos )
end function
function DQBmouseY () as integer 'OK
  _DQB_PumpEvents()
  _return( tDQB.tMouse.YPos )  
end function
function DQBmouseLB () as integer 'OK
  _DQB_PumpEvents()  
  _return( (tDQB.tMouse.Btn and 5)<>0 )  
end function
function DQBmouseRB () as integer 'OK
  _DQB_PumpEvents()  
  _return( (tDQB.tMouse.Btn and 6)<>0 )  
end function
sub DQBsetMousePos (byval x as integer, byval y as integer) 'OK
  _DQB_PumpEvents()  
  with tDQB.tMouse
    if x < .L then x=.L
    if y < .T then y=.T
    if x > .R then x=.R
    if y > .B then y=.B
    .XPos = x : .YPos = y
    if .bLock=0 then setmouse x,y
  end with
end sub
sub DQBmouseShow () 'OK
  _DQB_PumpEvents()  
  if TDQB.tMouse.bLock=0 then setmouse ,,1
end sub
sub DQBmouseHide () 'OK
  _DQB_PumpEvents()
  if TDQB.tMouse.bLock=0 then setmouse ,,0
end sub
sub DQBsetMouseRange (byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer) 'OK
  _DQB_PumpEvents()  
  with tDQB.tMouse
    if x1 > x2 then swap x1,x2
    if y1 > y2 then swap y1,y2
    if x1<0 orelse x2>319 orelse y1<0 orelse y2>199 then
      if tDQB.tMouse.bLock = 0 then
        tDQB.tMouse.bLock = 1 : setmouse ,,0,1
      end if
    else
      if tDQB.tMouse.bLock then
        tDQB.tMouse.bLock = 0 : setmouse ,,1,0
      end if
    end if
    .L=x1 : .T=y1 : .R=x2 : .B=y2
  end with
end sub
sub DQBsetMouseShape (hotX as integer, hotY as integer, Shape as string)
  Unimplemented()
end sub
sub DQBsetMouseSpeed (byval HorSpeed as integer, byval VerSpeed as integer)
  Unimplemented()
end sub
sub DQBresetMouse () 'OK
  with tDQB.tMouse
    if .bLock then .bLock = 0 
    .L=0 : .T=0 : .R=319 : .B=199
    .XPos = 160 : .YPos = 100
    setmouse .XPos,.YPos,1,0
  end with  
end sub

' Procedures from SOUND.OBJ:
function DQBinstallSB (byval VolActive as integer, byval Channels as integer, byval Freq as integer, byval BaseAddr as integer, byval IRQ as integer, byval DMA as integer) as integer
  UnimplementedN()
end function
function DQBloadSound (Slot as integer, FileName as string) as integer
  UnimplementedN()
end function
function DQBloadRawSound (Slot as integer, FileName as string, Offset as long, Length as long) as integer
  UnimplementedN()
end function
function DQBinUse (byval Voice as integer) as integer 'dummy
  _return(0)
end function
sub DQBplaySound (byval SoundNum as integer, byval Voice as integer, byval Freq as integer, byval LoopFlag as integer)
  Unimplemented()
end sub
sub DQBsetVoiceVol (byval Voice as integer, byval NewVol as integer)
  Unimplemented()
end sub
sub DQBpauseSound ()
  Unimplemented()
end sub
sub DQBresumeSound ()
  Unimplemented()
end sub
sub DQBstopVoice (byval Voice as integer)
  Unimplemented()
end sub
sub DQBremoveSB ()
  Unimplemented()
end sub
sub DQBsetVolume (byval Volume as integer)
  Unimplemented()
end sub

' Procedures from 3D.OBJ:
sub DQBtri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval x3 as integer, byval y3 as integer, byval Col as integer)
  Unimplemented()  
end sub
sub DQBbtri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval x3 as integer, byval y3 as integer, byval Col as integer, byval BMap as integer)
  Unimplemented()
end sub
sub DQBgtri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval c1 as integer, byval x2 as integer, byval y2 as integer, byval c2 as integer, byval x3 as integer, byval y3 as integer, byval c3 as integer)
  Unimplemented()
end sub
sub DQBbgtri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval c1 as integer, byval x2 as integer, byval y2 as integer, byval c2 as integer, byval x3 as integer, byval y3 as integer, byval c3 as integer, byval BMap as integer)
  Unimplemented()
end sub
sub DQBttri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval x3 as integer, byval y3 as integer, byval u1 as integer, byval v1 as integer, byval u2 as integer, byval v2 as integer, byval u3 as integer, byval v3 as integer, byval DummyTextureSeg as integer, byval TextureOff as any ptr) 'OK
  
  dim as any ptr pLayer : pLayer = tDQB.pLayer(Layer and 7)
  dim as long TexSize : TexSize = tDQB.iTexSize  
  
  'line pLayer,(x1,y1)-(x2,y2),63
  'line pLayer,-(x3,y3),63
  'line PLayer,-(x1,y1),63
  'exit sub
  
  #define _Min( _A , _B ) iif( (_A) < (_B) , (_A) , (_B) )
  #define _Max( _A , _B ) iif( (_A) > (_B) , (_A) , (_B) )

  ' Bounding box for triangle
  dim as long minX : minX = (_Min(x1, _Min(x2, x3)))
  dim as long minY : minY = (_Min(y1, _Min(y2, y3)))
  dim as long maxX : maxX = (_Max(x1, _Max(x2, x3)))
  dim as long maxY : maxY = (_Max(y1, _Max(y2, y3)))
  
  dim as double InvSize : InvSize = 1/TexSize
  dim as double fu1,fu2,fu3 : fu1=u1*InvSize : fu2=u2*InvSize : fu3=u3*InvSize
  dim as double fv1,fv2,fv3 : fv1=v1*InvSize : fv2=v2*InvSize : fv3=v3*InvSize
  
  dim as double denom : denom = 1/(((y2-clng(y3)) * (x1-x3) + (x3-clng(x2)) * (y1-y3)))
  #define SampleTexture( _u , _v , _pTex ) _pTex[ ((cint((_v)*TexSize) and (TexSize-1))*texSize) + (cint((_u)*TexSize) and (TexSize-1)) ]
  
  if minX < 0 then minX = 0
  if minY < 0 then minY = 0
  if MaxX > 319 then MaxX = 319
  if MaxY > 199 then MaxY = 199
  
  dim as double fuStart, fvStart, fuStep , fvStep 
  fuStep  = (y2-y3)*denom    : fvStep  = (y3-y1)*denom
  fuStart = (minX-x3)*fuStep : fvStart = (minX-x3)*fvStep
  
  dim as pixel ptr pTex,pTgt : pTex = TextureOff+4    
  dim as pixel uPix  
  if pLayer=0 then screenlock : pTgt = screenptr else pTgt = pLayer+4
  pTgt += 320l*minY
  
  dim as double fu,fv , texU,texV
  dim as long x,y
  for y = minY to maxY
    fu = fuStart + ((x3-clng(x2))*(y-y3))*denom
    fv = fvStart + ((x1-clng(x3))*(y-y3))*denom
    for x = minX to maxX          
      ' Check if the pixel is inside the triangle
      if fu >= 0 Andalso fv >= 0 then
        dim as single fw : fw = 1.0 - fu - fv
        if fw >= 0 then
          ' Interpolate UV coordinates
          texU = fu*fu1 + fv*fu2 + fw*fu3
          texV = fu*fv1 + fv*fv2 + fw*fv3                
          uPix = SampleTexture( texU,texV , pTex )
          if uPix then pTgt[x] = uPix
        end if
      end if
      ' update barycentric coordinates
      fu += fuStep : fv += fvStep
    next x
    pTgt += 320
  next y  
  
  'line pLayer, (x1,y1)-(x2,y2), 63 'point(u1 and (TexSize-1),v1 and (TexSize-1),pTex)
  'line pLayer, -(x3,y3), 63 'point(u2 and (TexSize-1),v2 and (TexSize-1),pTex)
  'line pLayer, -(x1,y1), 63 'point(u3 and (TexSize-1),v3 and (TexSize-1),pTex)
  
  if pLayer=0 then screenlock
  
end sub
sub DQBbttri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval x3 as integer, byval y3 as integer, byval u1 as integer, byval v1 as integer, byval u2 as integer, byval v2 as integer, byval u3 as integer, byval v3 as integer, byval DummyTextureSeg as integer, byval TextureOff as any ptr, byval BMap as integer)
  Unimplemented()
end sub
sub DQBfttri (byval Layer as integer, byval x1 as integer, byval y1 as integer, byval x2 as integer, byval y2 as integer, byval x3 as integer, byval y3 as integer, byval u1 as integer, byval v1 as integer, byval u2 as integer, byval v2 as integer, byval u3 as integer, byval v3 as integer, byval DummyTextureSeg as integer, byval TextureOff as any ptr)
  Unimplemented()
end sub
sub DQBsetTextureSize (byval Size as integer) 'OK
  dim as long iSz : iSz = 1
  while (iSz+iSz) <= Size
    iSz += iSz
  wend
  tDQB.iTexSize = iSz      
  if iSz <> Size then
    DbgPrint("Texture size=" & iSz & " (requested " & Size & ")")
  end if
end sub
sub xmsapi () 'OK (nothing)
  rem :)
end sub
sub XmsFree (byref fmem as long, byref tmem as long) 'OK
  LogFunc "XmsFree(&fmem , &tmem)"
  fmem = fre(-1) : tmem = fre(-1)
end sub
function movefromxms ( byval Kb as long , byval hHandle as integer , byval xmsoffs as long , byval DummySegm as integer , byval Offs as any ptr ) as integer 'OK
  LogFunc "movefromxms(Kb=" & Kb & " , hHandle=" & hHandle & " , xmsoffs=" & xmsoffs & " , Offs=0x" & hex$(Offs,8) & ")"
  if cuint(hHandle)>255 orelse tDQB.pXms(hHandle)=0 then DbgPrint("movefromxms: bad handle: " & hhandle) : _return(1)  
  fb_memcopy( *cptr(u8 ptr,Offs) , *cptr(u8 ptr,tDQB.pXms(hHandle)+xmsoffs) , Kb )
  _return(0) 
end function
function move2xms ( byval Kb as long , byval DummySegm as integer , byval Offs as any ptr , byval hHandle as integer , byval xmsoffs as long ) as integer 'OK
  LogFunc "move2xms(Kb=" & Kb & " , Offs=0x" & hex$(Offs,8) & " , hHandle=" & hHandle & " , xmsoffs=" & xmsoffs & ")"
  if cuint(hHandle)>255 orelse tDQB.pXms(hHandle)=0 then DbgPrint("movefromxms: bad handle: " & hhandle) : _return(1)
  fb_memcopy( *cptr(u8 ptr,tDQB.pXms(hHandle)+xmsoffs) , *cptr(u8 ptr,Offs) , Kb )
  _return(0)  
end function
function XmsdeAlloc (byval hHandle as integer) as integer 'OK
  LogFunc "XmsdeAlloc( hHandle=" & hHandle & ")"
  if cuint(hHandle)>255 then _return(0)
  if tDQB.pXms(hHandle) = 0 then _return(0)
  deallocate( tDQB.pXms(hHandle) )
  tDQB.pXms(hHandle) = 0
  _return( hHandle )
end function
function XmsAlloc (byval Kb as integer) as integer 'OK
  dim as long i
  for i = 1 to 255
    if tDQB.pXms(i)=0 then
      tDQB.pXms(i) = callocate( Kb*1024l ) 
      if tDQB.pXms(i) then 
        LogFunc i & "=XmsAlloc( Kb=" & kB & ")"
        _return( i )
      end if
      DbgPrint "(failed) XmsAlloc( Kb=" & kB & ")"
      _return(0)      
    end if
  next i
  DbgPrint "(failed) XmsAlloc( Kb=" & kB & ")"
  _return( 0 )
end function
function XmsVersion () as integer
  UnimplementedN()
end function
function XmsRealloc (byval hHandle as integer, byval Kb as integer) as integer 'OK
  LogFunc "XmsRealloc( hHandle=" & hHandle & " , Kb=" & kB & ")"
  if cuint(hHandle)>255 then _return(0)
  if tDQB.pXms(hHandle) = 0 then _return(0)
  dim pNew as any ptr
  pNew = reallocate( tDQB.pXms(hHandle) , Kb*1024 )
  if pNew then    
    tDQB.pXms(hHandle) = pNew
    _return(hHandle)
  end if
  _return(0)    
end function
